-- Во избежание конфликтов разместим все функции, обеспечивающие
-- партиционирование, в отдельной схеме.
do $$
begin
    if not exists (
        select schema_name
          from information_schema.schemata
          where schema_name = 'partitioning'
    )
    then
      execute 'create schema partitioning';
    end if;

END
$$;

create or replace function partitioning.getattr(
    rec anyelement,
    table_name text,
    column_name text
) returns text as $BODY$
    --
    -- Возвращает значение колонки column_name из записи rec
    -- (аналог getattr() в Python)
    --
    declare
        column_value text;
    begin
        execute 'select ($1::' || table_name || ').' || column_name  into column_value using rec;
        return column_value;
    end;
$BODY$ language plpgsql;


create or replace function partitioning.get_patrition_name(
    parent_table text,
    date_value timestamp
) returns text as $BODY$
    --
    -- Формирует название раздела из имени родительской таблицы и значения
    -- колонки, определяющей раздел таблицы.
    --
    declare
        month text;
        year text;
    begin
        month := 'm' || lpad(date_part('month', date_value)::TEXT, 2, '0');
        year := 'y' || date_part('year', date_value)::TEXT;
        return parent_table || '_' || year || month;
    end;
$BODY$ language plpgsql;


create or replace function partitioning.set_partition_constraint(
    partition_name text,
    column_name text,
    year integer,
    month integer
) returns void as $BODY$
    --
    -- Устанавливает ограничения для указанного раздела.
    -- Ранее установленные check-ограничения удаляются.
    --
    declare
        rec record;
        constraint_name text;
        min_timestamp timestamp with time zone;
        max_timestamp timestamp with time zone;
    begin
        constraint_name := partition_name || '_date_check';

        -- Удаление имеющихся check-ограничений.
        for rec in
            select conname
            from pg_constraint
            where
                contype = 'c' and  -- тип ограничения: check
                conname like '%_date_check' and
                conrelid = partition_name::regclass::oid
        loop
            execute
                'alter table ' || partition_name || ' ' ||
                    'drop constraint ' || rec.conname;
        end loop;

        -- Определение граничных значений для раздела.
        min_timestamp := format('%s-%s-01 00:00:00', year, month)::timestamp;
        max_timestamp := min_timestamp + interval '1 month'
                         - interval '1 second';

        execute
            'alter table ' || partition_name || ' ' ||
                'add constraint ' || partition_name || '_date_check ' ||
                'check (' ||
                    column_name || ' between ' ||
                    '''' || min_timestamp ||
                    ''' and ''' ||
                    max_timestamp || '''' ||
                ')';
    end;
$BODY$ language plpgsql;


create or replace function partitioning.table_exists(
    name text
) returns boolean as $BODY$
    --
    -- Проверяет, существует ли в БД таблица table_name.
    --
    begin
        return exists(
            select 1
            from information_schema.tables
            where table_name = name
        );
    end;
$BODY$ language plpgsql;


create or replace function partitioning.create_partition(
    parent_table_name text,
    pk_column_name text,
    column_name text,
    column_value date
) returns void as $BODY$
    --
    -- Создает раздел таблицы.
    --
    declare
        partition_name text;
    begin
        -- имя создаваемой таблицы
        partition_name := partitioning.get_patrition_name(
            parent_table_name, column_value
        );

        -- создание раздела
        execute
            'create table if not exists ' || partition_name || ' '
            '(' ||
                'like ' || parent_table_name || ' ' ||
                'including defaults ' ||
                'including constraints ' ||
                'including indexes' ||
            ') ' ||
            'inherits (' || parent_table_name || ')';

        perform partitioning.set_partition_constraint(
            partition_name,
            column_name,
            date_part('year', column_value)::integer,
            date_part('month', column_value)::integer
        );

        -- создание триггера для раздела обрабатывающего изменения, которые
        -- требуют переноса записи из одного раздела таблицы в другой
        execute
            'create trigger partitioning__' || partition_name || '__update '
                'before update on ' || partition_name || ' ' ||
                'for each row ' ||
                'execute procedure partitioning.before_update(' ||
                    '''' || pk_column_name || ''', ''' || column_name || ''''
                ')';
    end;
$BODY$ language plpgsql;


create or replace function partitioning.before_insert()
returns trigger as $BODY$
    --
    -- Обработчик добавления записей в партиционированную таблицу. При
    -- отсутствии необходимого раздела создает его и вставляет в него
    -- добавляемую запись.
    --
    declare
        partition_name text;  -- имя раздела
        pk_column_name text;  -- имя колонки с первичным ключом
        column_name text;  -- имя колонки, определяющей раздел таблицы
        column_value date;  -- значение поля, определяющее раздел таблицы
    begin
        pk_column_name := TG_ARGV[0];
        column_name := TG_ARGV[1];

        column_value := partitioning.getattr(NEW, TG_TABLE_NAME, column_name);
        partition_name := partitioning.get_patrition_name(
            TG_TABLE_NAME, column_value
        );

        if not partitioning.table_exists(partition_name) then
            -- целевого раздела нет, создадим его
            perform partitioning.create_partition(
                TG_TABLE_NAME, pk_column_name, column_name, column_value
            );
        end if;

        -- добавим запись в целевой раздел
        execute
            'insert into ' || partition_name || ' ' ||
            'values (($1).*);'
            USING NEW;

        -- в родительскую таблицу добавлять записи не нужно
        return NEW;
    end;
$BODY$ language plpgsql;


create or replace function partitioning.after_insert()
returns trigger as $BODY$
    --
    -- Удаляет из родительской таблицы добавленную запись.
    -- Это необходимо в связи с тем, что создание записи в Django выполняется с
    -- параметром RETURNING b и без реального создания записи в родительской
    -- таблице значение ключевого поля будет равно NULL.
    --
    declare
        pk_column_name text;  -- имя колонки с первичным ключом
    begin
        pk_column_name := TG_ARGV[0];

        execute
            'delete from only ' || TG_TABLE_NAME || ' ' ||
            'where ' || pk_column_name || ' = ' ||
                partitioning.getattr(NEW, TG_TABLE_NAME, pk_column_name);

        -- в родительскую таблицу добавлять записи не нужно
        return NEW;
    end;
$BODY$ language plpgsql;


create or replace function partitioning.before_update()
returns trigger as $BODY$
    --
    -- Обработчик изменения записей в разделах таблицы. В случае изменения
    -- в записи значения, определяющего раздел таблицы, переносит эту запись
    -- в другой раздел. При необходимости создает новый раздел таблицы.
    --
    declare
        parent_table_name text;  -- имя родительской таблицы
        dst_table_name text;  -- имя раздела, в который будет перенесена запись
        pk_column_name text;  -- имя колонки с первичным ключом
        column_name text;  -- имя колонки, определяющей раздел таблицы
        old_column_value date;  -- исходное значение колонки, определяющее
                               -- раздел таблицы
        new_column_value date;  -- новое значение колонки, определяющей раздел
                               -- таблицы
    begin
        pk_column_name := TG_ARGV[0];
        column_name := TG_ARGV[1];

        old_column_value := date_trunc(
            'month', partitioning.getattr(OLD, TG_TABLE_NAME, column_name)::date
        );
        new_column_value := date_trunc(
            'month', partitioning.getattr(NEW, TG_TABLE_NAME, column_name)::date
        );

        -- определение имени родительской таблицы
        if TG_TABLE_NAME ~  '_y([0-9]{4})m([0-9]{2})$' then
            -- запись находится в разделе таблицы
            parent_table_name := left(
                TG_TABLE_NAME, length(TG_TABLE_NAME) - length('_yXXXXmXX')
            );
        else
            -- запись находится в родительской таблице
            parent_table_name := TG_TABLE_NAME;
        end if;

        dst_table_name := partitioning.get_patrition_name(
            parent_table_name, new_column_value
        );

        if TG_TABLE_NAME = dst_table_name then
            -- перенос записи в другой раздел не требуется
            return NEW;
        end if;

        -- значение поля, определяющего раздел таблицы, изменено так, что
        -- необходим перенос записи в другой раздел

        if not partitioning.table_exists(dst_table_name) then
            -- целевого раздела нет, создадим его
            perform partitioning.create_partition(
                parent_table_name, pk_column_name, column_name, new_column_value
            );
        end if;

        -- перенос записи в другой раздел таблицы
        execute
            'delete from ' || TG_TABLE_NAME || ' ' ||
            'where ' || pk_column_name || ' = ' ||
                partitioning.getattr(NEW, TG_TABLE_NAME, pk_column_name);
        execute
            'insert into ' || dst_table_name || ' ' ||
            'values (($1).*)' using NEW;

        return null;
    end;
$BODY$ language plpgsql;
